// x10broker.cpp
//
//	2005/09/01
//  Tomoichi Ebata
//  ebata@ff.iij4u.or.jp
//

#include	"stdafx.h"
#include	<stdio.h>
#include	<windows.h>
#include	<winsock.h>
#include	<process.h>
#include    <stdlib.h>
#include	"x10broker.h"

CRITICAL_SECTION	g_Lock;  

void Systemlog(char *msg)
{
	FILE *f;

	if ((f = fopen(LOG_FILE, "a")) != NULL)
	{
		fprintf(f,"SYSTEMLOG:");
		fprintf(f,msg);
		fprintf(f,"\n");
		fclose(f);
	}

}

void Errorlog(char *msg)
{
	FILE *f;

	if ((f = fopen(LOG_FILE, "a")) != NULL)
	{
		fprintf(f,"ERRORLOG:");
		fprintf(f,msg);
		fprintf(f,"\n");
		fclose(f);
	}

}

X10EventBroker::X10EventBroker(
							   HANDLE cph, 
							   int lp)
{
	comPortHandle = cph;
	serverPort = lp;

	WSADATA			Data;
	int status = WSAStartup(MAKEWORD(1,1), &Data);
	if (status != 0)
		throw SocketPanic(0,"ERROR:WSAStartup unsuccessful\n");

	SOCKADDR_IN		serverSockAddr;
	memset(&serverSockAddr, 0, sizeof(serverSockAddr));

	serverSockAddr.sin_port = htons(serverPort);
	serverSockAddr.sin_family = AF_INET;
	serverSockAddr.sin_addr.s_addr = htonl(INADDR_ANY);

	listenSocket=socket(AF_INET, SOCK_STREAM, 0);
	if(listenSocket == INVALID_SOCKET)
	{
		throw SocketPanic(0,"ERROR:socket unsuccessful in TCPTunnelServer\n");
	}

	status = bind(
		listenSocket, (LPSOCKADDR)&serverSockAddr, sizeof(serverSockAddr));
	if(status == SOCKET_ERROR)
	{
		throw SocketPanic(0,"ERROR:bind unsuccessful in TCPTunnelServer\n");
	}
	
	status = listen(listenSocket, 1);
	if(status == SOCKET_ERROR)
	{
		throw SocketPanic(0, "ERROR:listen unsuccessful in TCPTunnelServer\n");
	}
}


X10EventBroker::start()
{

	SOCKADDR_IN		clientSockAddr;
	int				addrLen=sizeof(SOCKADDR_IN);
	
	while(1)
	{
		clientSocket 
			= accept(listenSocket,(LPSOCKADDR)&clientSockAddr, &addrLen);
		if (clientSocket == INVALID_SOCKET)
		{
			shutdown(clientSocket,2);
			closesocket(clientSocket);
			
			throw SocketPanic(0,"ERROR: accept unsuccessful\n");
		}

		char recvData[MAXBUFLEN];

		int numberRecv = recv(clientSocket, recvData, MAXBUFLEN, 0);
		if ((numberRecv == SOCKET_ERROR) || (numberRecv == 0))
		{	
			lifeFlag = false;
			break;
		}

		closesocket(clientSocket);

		DWORD numWrite;

		EnterCriticalSection(&g_Lock);
		WriteFile(comPortHandle, recvData, numberRecv, &numWrite,0);
		LeaveCriticalSection(&g_Lock);
	}
}

X10EventBroker::end()
{
	// revisit
}

unsigned int __stdcall _x10EventBrokerThread(void *eventBrokerThreadInfo)
{
	struct EventBrokerThreadInfo *ebti 
		= (struct EventBrokerThreadInfo *)eventBrokerThreadInfo;

	X10EventBroker
		x10EventBroker(ebti->comPortHandle, ebti->serverPort);

	try
	{
		x10EventBroker.start();
	}
	catch (SocketPanic socketPanic) 
	{
		printf("%s\n",socketPanic.reason());
	}

	x10EventBroker.end();

	return(0);
}



X10EventListener::X10EventListener(
								   HANDLE cph, 
								   struct PolicyInfo pi)
{
	comPortHandle = cph;
	memcpy(&policyInfo, &pi, sizeof(struct PolicyInfo));
}

X10EventListener::parse_and_execute(struct ControlCommand &cc)
{
#if _LOG
	Systemlog("Start parse_and_execute\n");
#endif

	for(int i = 0 ; i < policyInfo.policyInfoCount ; i++)
	{
		if (strncmp(policyInfo.policyScript[i].conditionInfo.addressCommand , cc.addressCommand, 4) == 0)
		{
			if (strncmp(policyInfo.policyScript[i].conditionInfo.actionCommand, cc.actionCommand, 4) == 0)
			{
				STARTUPINFO sinfo;
				PROCESS_INFORMATION pinfo;
					
				ZeroMemory( &sinfo, sizeof( STARTUPINFO ) );
				ZeroMemory( &pinfo, sizeof( PROCESS_INFORMATION ) );
				sinfo.cb = sizeof( STARTUPINFO );
				sinfo.dwFlags = STARTF_USESHOWWINDOW;
				sinfo.wShowWindow = SW_SHOWNORMAL;

#if _LOG
				Sysmtelog(policyInfo.policyScript[i].actionScript);
#endif
				if ( !CreateProcess( NULL,
					policyInfo.policyScript[i].actionScript,
					NULL,
					NULL,
					FALSE,
					0,
					NULL,
					NULL,
					&sinfo,
					&pinfo))
				{
					Errorlog("Fail to CreateProcess, GetLastError ->");

					char dummy[10];
					itoa(GetLastError(), dummy,10);
					Errorlog(dummy);				
				}

			}
		}
	}
}



X10EventListener::start()
{
	int flag;
	struct ControlCommand cc;

//	DWORD mask;
//	SetCommMask(comPortHandle, EV_RXFLAG);	

	while(1)
	{
		flag = 0;
		while(1)
		{
			Sleep(10000);	// X10 interface should be checked every 10 second

//			WaitCommEvent(comPortHandle, &mask, 0);
						
			char str[100];
			memset(str,0 ,sizeof(str));
			DWORD	numRead;

			EnterCriticalSection(&g_Lock);
			if(!ReadFile(comPortHandle, str, sizeof(str), &numRead, 0))
			{
				throw FilePanic(0,"....ReadFile");
				break;
			}
			LeaveCriticalSection(&g_Lock);	
			
			int strlen_str = strlen(str);
			for(int i = 0; i < strlen_str ; i++)
			{
				// When acction code(like "!A20" "!A28") is received, send both 
				// the action code and a address code to the X10 interface.
				// The address code must be recieved before receiving an action code.

				// The following '4' menas a length of an action code and an address code
				// lile "&A01", "!A20"
				
				if (str[i] == '!')
				{
					strncpy(cc.actionCommand, str + i, 4);
					parse_and_execute(cc);
					i = i + 4;
				}
				if (str[i] == '&')
				{
					strncpy(cc.addressCommand, str + i, 4);
					i = i + 4;
				}
			}
		}
	}
}

X10EventListener::end()
{
	// revisit
}
					

unsigned int __stdcall _x10EventListenerThread(void *eventListenerThreadInfo)
{
	struct EventListenerThreadInfo *elti 
		= (struct EventListenerThreadInfo *)eventListenerThreadInfo;

	X10EventListener
		x10EventListener(elti->comPortHandle,elti->policyInfo);

	try
	{
		x10EventListener.start();
	}
	catch (SocketPanic socketPanic) 
	{
		printf("%s\n",socketPanic.reason());
	}

	x10EventListener.end();

	return(0);
}

#if _NTS
DWORD ServiceThread(LPDWORD param)  // This is for NT Service
#else
int main(void) 
#endif
{
	InitializeCriticalSection(&g_Lock);
	
	struct EventListenerThreadInfo	eventListenerThreadInfo;
	struct EventBrokerThreadInfo	eventBrokerThreadInfo;

	FILE *f;
	char string[1000];
	int counter = 0;

	char buffer[MAX_PATH];

	char comPort[4];

	//// Getting Policy Information ////
	try
	{
		DWORD len = GetCurrentDirectory(MAX_PATH, buffer);
		
		if (len == 0) 
		{
			throw FilePanic(0,"Could not get current Directory...\n");
		}

	
		char configFileName[MAX_PATH];
		strcpy(configFileName, buffer);
		strcat(configFileName, "\\x10policy.cfg");


		f = fopen(configFileName,"r");
		if (!f)
		{
			strcat(configFileName, " is not found.\n");
			throw FilePanic(0,configFileName);
		}
		
		while((string[counter++] = fgetc(f)) != EOF);

		counter--;
		string[counter] = NULL;
		fclose(f);

		//if (counter > 30)
		//{
		//	throw FilePanic(0,"Stopped program because of more than 30 ports described in tcptunnel.cfg\n"); 
		//}
	}
	catch (FilePanic filePanic) 
	{
		printf("%s\n",filePanic.reason());
		exit(0);
	}

	char seps[]   = "\t\n";
	char *token;

//	unsigned threadID;
	struct PolicyInfo policyInfo;

	token = strtok( string, seps );

	policyInfo.policyInfoCount = 0;
	
	try
	{
		bool flag = true;

		while( token != NULL )
		{
			strcpy(policyInfo.policyScript[policyInfo.policyInfoCount].conditionInfo.addressCommand, token);

			if ((token = strtok( NULL, seps )) != NULL)
				strcpy(policyInfo.policyScript[policyInfo.policyInfoCount].conditionInfo.actionCommand, token);
			else
			{
				flag = false;
				break;
			}

			if ((token = strtok( NULL, seps )) != NULL)
			strcpy(policyInfo.policyScript[policyInfo.policyInfoCount].actionScript,token);
			else
			{
				flag = false;
				break;
			}

			token = strtok( NULL, seps );
			policyInfo.policyInfoCount++;
		}
		if ( flag == false)
			throw FilePanic(0,"Unable to make more than 30 policy information");
	}
	catch (FilePanic filePanic) 
	{
		printf("%s\n",filePanic.reason());
		exit(0);
	}

	memcpy(&(eventListenerThreadInfo.policyInfo), &policyInfo, sizeof(struct PolicyInfo));
	
	try
	{
		if ( policyInfo.policyInfoCount >= 30 )
		{
			throw FilePanic(0,"Unable to make more than 30 policy information");
		}
	}
	catch (FilePanic filePanic)
	{
		printf("%s\n",filePanic.reason());
		exit(0);
	}

	//// Getting Communication Information ////

	memset(string, NULL, sizeof(string));
	counter = 0;

	try
	{
		DWORD len = GetCurrentDirectory(MAX_PATH, buffer);
		
		if (len == 0) 
		{
			throw FilePanic(0,"Could not get current Directory...\n");
		}
	
		char configFileName[MAX_PATH];
		strcpy(configFileName, buffer);
		strcat(configFileName, "\\x10broker.cfg");		

		f = fopen(configFileName,"r");
		if (!f)
		{
			strcat(configFileName, " is not found.\n");
			throw FilePanic(0,configFileName);
		}
		
		while((string[counter++] = fgetc(f)) != EOF);

		counter--;
		string[counter] = NULL;
		fclose(f);

	}
	catch (FilePanic filePanic) 
	{
		printf("%s\n",filePanic.reason());
		exit(0);
	}

	strcpy(seps, "\t\n");

//	struct ConnectInfo connectInfo;
	token = strtok( string, seps );

	int connectInfoCounter = 0;
	
	try
	{
		bool flag = true;

		while( token != NULL )
		{
			if ((strcmp(token,"serverport") == 0) || (strcmp(token,"SERVERPORT") == 0))
			{
				printf("%s\t",token);
				if ((token = strtok( NULL, seps )) == NULL)
					flag = false;

				printf("%s\t",token);
				eventBrokerThreadInfo.serverPort = atoi(token);

			}
			else if ((strcmp(token,"com") == 0) || (strcmp(token,"COM") == 0))
			{
				if ((token = strtok( NULL, seps )) == NULL)
					flag = false;
				printf("%s\t",token);

				strcpy(comPort, token);
			}
			
			if ((token = strtok( NULL, seps )) == NULL)
				break;
		}
	}
	
	catch (FilePanic filePanic) 
	{
		printf("%s\n",filePanic.reason());
		exit(0);
	}

	//// Com port initialize /////
	
	// Open the comm port. 
	HANDLE comHandle;
	
	comHandle = CreateFile(
		comPort, 
		GENERIC_READ|GENERIC_WRITE, 
		0, 
		0, 
		OPEN_EXISTING, 
		FILE_ATTRIBUTE_NORMAL, 
		0);

	int dummy = GetLastError();

	if (comHandle == INVALID_HANDLE_VALUE)
		throw ComPanic(GetLastError(),"...CreateFile -> check ComPort Number");

	COMMTIMEOUTS timeouts;
	timeouts.ReadIntervalTimeout = MAXDWORD;//Non-blocking mode //0: blocking mode
	timeouts.ReadTotalTimeoutMultiplier = 0;
	timeouts.ReadTotalTimeoutConstant = 0;
	timeouts.WriteTotalTimeoutMultiplier = 0;
	timeouts.WriteTotalTimeoutConstant = 0;
	SetCommTimeouts( comHandle, &timeouts );

	// Get the current settings of the COMM port 
	BOOL success;
	DCB dcb;
	
	success = GetCommState(comHandle, &dcb);
	if (!success) 
		throw ComPanic(GetLastError(),"...GetCommState");

   // Modify the baud rate, etc.
	dcb.BaudRate = 9600;
	dcb.ByteSize = 8;
	dcb.Parity = NOPARITY;
	dcb.StopBits = ONESTOPBIT;
	dcb.EvtChar = '\n';		//@Cxg̐ݒiReturn keyj

	// Apply the new comm port settings
	success = SetCommState(comHandle, &dcb);
	if (!success) 
		throw ComPanic(GetLastError(),"...SetCommState");

	// Set the Data Terminal Ready line
	EscapeCommFunction(comHandle, SETDTR);
	
	//// Set X10 interface as a x10 promiscous mode
	char str[10];
	unsigned long numWrite;

	memset(str,0,sizeof(str));	
	strcpy(str, "*E1");  // X10 promicous mode

	try
	{
		if(!WriteFile(comHandle, str, strlen(str) , &numWrite, 0))
		{
			throw FilePanic(0,"ERROR:WriteFile unsuccessful\n");
		}
	}
	catch (FilePanic filePanic) 
	{
		printf("%s\n",filePanic.reason());
	}

	eventListenerThreadInfo.comPortHandle = eventBrokerThreadInfo.comPortHandle
		= comHandle;

	//// Make X10EventBroker and X10EventListener thread ////
	HANDLE threadHandle[2];
	unsigned threadID;

	threadHandle[0]
			= (HANDLE)_beginthreadex(
			NULL, 0, &_x10EventListenerThread, &eventListenerThreadInfo, 0, &threadID); 

	threadHandle[1]
			= (HANDLE)_beginthreadex(
			NULL, 0, &_x10EventBrokerThread, &eventBrokerThreadInfo, 0, &threadID); 

	WaitForMultipleObjects(2, threadHandle, TRUE,INFINITE);

	for(int i = 0; i < 2; i++)
		CloseHandle(threadHandle[i]);

	DeleteCriticalSection(&g_Lock);

	return 0;
}
